#!/usr/bin/env python3

############################################################################
#Script Name    : S004007_DVFS_cpufreq.py
#Description    : Sript for testing CPU frequency (DVFS)
#                 Covered test cases: C490488
#Author         : Joying Kuo
#Email          : joying.kuo@mediatek.com
###########################################################################

import os, sys
import unittest
import subprocess
import time

BASE = '/sys/devices/system/cpu'

def get_soc_name():
    ret = subprocess.run(["cat /proc/device-tree/compatible"],shell=True,stdout=subprocess.PIPE,stderr=subprocess.PIPE,encoding="utf-8")
    soc = ret.stdout.split(',')[1].split('-')[0]
    return soc
soc = get_soc_name()

if soc == 'mt8365':
    FREQS1 = '850000 918000 987000 1056000 1125000 1216000 1308000 1400000 1466000 1533000 1633000 1700000 1767000 1834000 1917000 2001000'
else:
    FREQS1 = '2000000 1903000 1807000 1710000 1614000 1517000 1421000 1325000 1228000 1132000 1035000 939000 842000 746000 650000 500000'

FREQS4 = '2200000 2093000 1986000 1880000 1773000 1666000 1560000 1453000 1346000 1240000 1133000 1026000 920000 813000 706000 600000'
FREQS6 = '2200000 2096000 1993000 1890000 1786000 1683000 1580000 1476000 1373000 1270000 1166000 1063000 960000 856000 753000 650000'

def read_attr(attr):
    path = os.path.join(BASE, attr)
    try:
        f = open(path)
        tmp = f.read().strip()
        f.close()
    except IOError:
        return ''

    return tmp

def read_attr_num(attr):
    tmp = read_attr(attr)
    if not tmp:
        return -1

    return int(tmp)

def write_attr(attr, value):
    path = os.path.join(BASE, attr)
    try:
        f = open(path, 'w')
        f.write(value)
        f.close()
    except IOError:
        return False

    return True

def read_cpus():
    cpus = []

    i = 0
    while True:
        t = read_attr('cpu%d/online' % i)
        if not t:
            break

        cpus.append(t)
        i += 1

    return cpus

def read_policies():
    policies = []
    i = 0
    while True:
        t = read_attr('cpufreq/policy%d/related_cpus' % i)
        if not t:
            break
        
        tmp = {}
        tmp['cpus'] = len(t.split(' '))
        policies.append(t)
        i += tmp['cpus']

    return policies

def read_pol_attr(policy, attr):
    return read_attr('cpufreq/policy%d/%s' % (policy, attr))

def read_pol_attr_num(policy, attr):
    return read_attr_num('cpufreq/policy%d/%s' % (policy, attr))

def write_pol_attr(policy, attr, value):
    return write_attr('cpufreq/policy%d/%s' % (policy, attr), value)

def stress_cpus():
    cpus = len(read_cpus())

    procs = []
    cmd = ["dd", "if=/dev/zero", "of=/dev/null"]
    for i in range(cpus):
        p = subprocess.Popen(cmd)
        procs.append(p)

    return procs

def stop_stress_cpus(procs):
    for p in procs:
        p.terminate()
        p.wait()

class TestCpuFreq(unittest.TestCase):

    def setUp(self):
        self.procs = []

    def test_sysfs_existence(self):
        self.assertTrue(os.path.isdir(BASE))

    def test_read_cpus(self):
        cpus = read_cpus()
        if soc == 'mt8365':
            self.assertEqual(len(cpus), 4)
        elif soc == 'mt8370':
            self.assertEqual(len(cpus), 6)
        else:
            self.assertEqual(len(cpus), 8)


    def test_read_policies(self):
        policies = read_policies()
        if soc == 'mt8365':
            self.assertEqual(len(policies), 1)
        else:
            self.assertEqual(len(policies), 2)
            
        self.assertEqual(read_pol_attr(0, 'scaling_available_frequencies'), FREQS1)
        
        if soc == 'mt8390':
            self.assertEqual(read_pol_attr(6, 'scaling_available_frequencies'), FREQS6)
        elif soc == 'mt8370':
            self.assertEqual(read_pol_attr(4, 'scaling_available_frequencies'), FREQS6)   
        elif soc =='mt8395' or soc == 'mt8195':
            self.assertEqual(read_pol_attr(4, 'scaling_available_frequencies'), FREQS4)    

    def test_performance_governor(self):
        gov0 = read_pol_attr(0, 'scaling_available_governors')
        self.assertTrue('performance' in gov0)
        tmp0 = read_pol_attr(0, 'scaling_governor')
        
        # policy 0
        max_freq = read_pol_attr(0, 'scaling_max_freq')
        self.assertTrue(write_pol_attr(0, 'scaling_governor', 'performance'))
        time.sleep(5)
        cur_freq = read_pol_attr(0, 'scaling_cur_freq')
        self.assertEqual(cur_freq, max_freq)
        self.assertTrue(write_pol_attr(0, 'scaling_governor', tmp0))

        
        if soc == 'mt8390':
            gov6 = read_pol_attr(6, 'scaling_available_governors')
            self.assertTrue('performance' in gov6)
            tmp6 = read_pol_attr(6, 'scaling_governor')
            
            # policy 6
            max_freq = read_pol_attr(6, 'scaling_max_freq')
            self.assertTrue(write_pol_attr(6, 'scaling_governor', 'performance'))
            time.sleep(5)
            cur_freq = read_pol_attr(6, 'scaling_cur_freq')
            self.assertEqual(cur_freq, max_freq)
            self.assertTrue(write_pol_attr(6, 'scaling_governor', tmp6))

        elif soc == 'mt8395' or soc == 'mt8195' or soc == 'mt8370':
            gov4 = read_pol_attr(4, 'scaling_available_governors')
            self.assertTrue('performance' in gov4)
            tmp4 = read_pol_attr(4, 'scaling_governor')
            
             # policy 4
            max_freq = read_pol_attr(4, 'scaling_max_freq')
            self.assertTrue(write_pol_attr(4, 'scaling_governor', 'performance'))
            time.sleep(5)
            cur_freq = read_pol_attr(4, 'scaling_cur_freq')
            self.assertEqual(cur_freq, max_freq)
            self.assertTrue(write_pol_attr(4, 'scaling_governor', tmp4))

    def test_schedutil_governor(self):
        gov0 = read_pol_attr(0, 'scaling_available_governors')
        self.assertTrue('schedutil' in gov0)
        tmp0 = read_pol_attr(0, 'scaling_governor')
        self.assertTrue(write_pol_attr(0, 'scaling_governor', 'schedutil'))
        
        # Add sleep to avoid spike on CPU loading
        time.sleep(5)
        prev_freq0 = read_pol_attr_num(0, 'scaling_cur_freq')
        
        
        if soc == 'mt8390':
            gov6 = read_pol_attr(6, 'scaling_available_governors')
            self.assertTrue('schedutil' in gov6)
            tmp6 = read_pol_attr(6, 'scaling_governor')
            self.assertTrue(write_pol_attr(6, 'scaling_governor', 'schedutil'))
            
            # Add sleep to avoid spike on CPU loading
            time.sleep(5)
            prev_freq6 = read_pol_attr_num(6, 'scaling_cur_freq')
            
        if soc == 'mt8395' or soc == 'mt8195' or soc == 'mt8370':
            gov4 = read_pol_attr(4, 'scaling_available_governors')
            self.assertTrue('schedutil' in gov4)
            tmp4 = read_pol_attr(4, 'scaling_governor')
            self.assertTrue(write_pol_attr(4, 'scaling_governor', 'schedutil'))
            
            # Add sleep to avoid spike on CPU loading
            time.sleep(5)
            prev_freq4 = read_pol_attr_num(4, 'scaling_cur_freq')
        
        self.procs = stress_cpus()
        time.sleep(10)
          
        # policy 0
        cur_freq = read_pol_attr_num(0, 'scaling_cur_freq')
        self.assertTrue(cur_freq > prev_freq0, 'cur: %d, prev: %d' % (cur_freq, prev_freq0))
        self.assertTrue(write_pol_attr(0, 'scaling_governor', tmp0))
    
        if soc == 'mt8390':
            # policy 6
            cur_freq = read_pol_attr_num(6, 'scaling_cur_freq')
            self.assertTrue(cur_freq > prev_freq6, 'cur: %d, prev: %d' % (cur_freq, prev_freq6))
            self.assertTrue(write_pol_attr(6, 'scaling_governor', tmp6))
        if soc == 'mt8395'  or soc == 'mt8195' or soc == 'mt8370':
            # policy 4
            cur_freq = read_pol_attr_num(4, 'scaling_cur_freq')
            self.assertTrue(cur_freq > prev_freq4, 'cur: %d, prev: %d' % (cur_freq, prev_freq4))
            self.assertTrue(write_pol_attr(4, 'scaling_governor', tmp4))
            

    def tearDown(self):
        stop_stress_cpus(self.procs)


if __name__ == '__main__':
    unittest.main()
